/*
Title:			peggy.c 
Date Created:   1/16/08
Last Modified:  1/16/08
Target:			Atmel ATmega164P 
Environment:	AVR-GCC
Purpose: Drive multiplexed or charlieplexed LED array

Fuse settings: 8 MHz internal RC oscillator, clock prescaler turned off (divide by 1)



Written by Windell Oskay, http://www.evilmadscientist.com/

Copyright 2008 Windell H. Oskay
Distributed under the terms of the GNU General Public License, please see below.

Additional license terms may be available; please contact us for more information.



More information about this project is at 
http://www.evilmadscientist.com/article.php/peggy



-------------------------------------------------
USAGE: How to compile and install



A makefile is provided to compile and install this program using AVR-GCC and avrdude.

To use it, follow these steps:
1. Update the header of the makefile as needed to reflect the type of AVR programmer that you use.
2. Open a terminal window and move into the directory with this file and the makefile.  
3. At the terminal enter
		make clean   <return>
		make all     <return>
		make install <return>
4. Make sure that avrdude does not report any errors.  If all goes well, the last few lines output by avrdude
should look something like this:

avrdude: verifying ...
avrdude: XXXX bytes of flash verified

avrdude: safemode: lfuse reads as E2
avrdude: safemode: hfuse reads as D9
avrdude: safemode: efuse reads as FF
avrdude: safemode: Fuses OK

avrdude done.  Thank you.


If you a different programming environment, make sure that you copy over 
the fuse settings from the makefile.


-------------------------------------------------

This code should be relatively straightforward, so not much documentation is provided.  If you'd like to ask 
questions, suggest improvements, or report success, please use the evilmadscientist forum:
http://www.evilmadscientist.com/forum/

-------------------------------------------------


Revision hitory:

1/22/2008 

Initial version
-------------------------------------------------


 This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


*/









#include <avr/io.h> 
#include <avr/interrupt.h>


/*
Need to add:

ADC routine: Darkness detection
Sleep ("OFF") Mode
*/
void delayLong()
{
unsigned int delayvar;
	delayvar = 0; 
	while (delayvar <=  50U)		
		{ 
			asm("nop");  
			delayvar++;
		} 
}




void dc() // Delay for Charlieplexing
{
unsigned int delayvar;

			delayvar = 0; 
	while (delayvar <=  100U)		
		{ 
			asm("nop");  
			asm("nop");  
			asm("nop");  
			asm("nop");  
			asm("nop");  
			delayvar++;
		} 

//Uncomment the following section to watch the charlieplexed display
//do its stuff, one line at a time.

/*
			delayvar = 0; 
	while (delayvar <=  65000U)		
		{ 
			asm("nop");  
			asm("nop");  
			asm("nop");  
			asm("nop");  
			asm("nop");  
			delayvar++;
		} 


			delayvar = 0; 
	while (delayvar <=  65000U)		
		{ 
			asm("nop");  
			asm("nop");  
			asm("nop");  
			asm("nop");  
			asm("nop");  
			delayvar++;
		} 
		
			delayvar = 0; 
	while (delayvar <=  65000U)		
		{ 
			asm("nop");  
			asm("nop");  
			asm("nop");  
			asm("nop");  
			asm("nop");  
			delayvar++;
		} 
		
		
			delayvar = 0; 
	while (delayvar <=  65000U)		
		{ 
			asm("nop");  
			asm("nop");  
			asm("nop");  
			asm("nop");  
			asm("nop");  
			delayvar++;
		} 
		
		
			delayvar = 0; 
	while (delayvar <=  65000U)		
		{ 
			asm("nop");  
			asm("nop");  
			asm("nop");  
			asm("nop");  
			asm("nop");  
			delayvar++;
		} 
		*/

}



SIGNAL(SIG_PIN_CHANGE0)
{ 

		
DDRD  = 0; // All D Inputs
PORTD = 0;

DDRC  = 0; // All C Inputs
PORTC = 0;

DDRB  = 0; // All B Inputs
PORTB = 0;

DDRA  = 0; // All A Inputs
PORTA = 0;

ADCSRA = 0; // Kill ADC

SMCR = 5U;
asm("sleep");		//Go to sleep!
}





int main (void)
{ 
asm("cli");		// DISABLE global interrupts


unsigned int ADIn,  ADoutTemp;
unsigned int ADCcount;

unsigned short charlieplexed, LightsOn, EnableLights, AutoMode; 

unsigned int AutoStage;
unsigned int j;		// General purpose indexing dummy variable

unsigned long iteration;

ADCcount = 0;
EnableLights = 1;	// Basic conditions met
LightsOn = 1;		// actually turn lights on
AutoMode = 1;   //Whether we are in the auto/start-up/power down sequence
AutoStage = 0;	// If so, which stage we are in
// 0: Initialization
// 1: Initial fading/blinking on
// 2: steady on for some hours
// 3: power down until it goes dark again

DDRA  = 0; // All A Inputs
PORTA = 0;



	
ADIn = 10000U;

// General Hardware Initialization:

MCUCR |= (1 << 4); // Disable pull-ups

PCMSK0 = 4U; // Pin change interrupt mask
PCICR = 1; // Enable PCINTs on PCMSK0

		
DDRD  = 0; // All D Inputs
PORTD = 255U;

DDRC  = 0; // All C Inputs
PORTC = 255U;

DDRB  = 0; // All B Inputs
PORTB = 255U;

DDRA  = 0; // All A Inputs
PORTA = 255U;




ADMUX = 5U;		// Channel 3 only

ADCSRA = _BV(ADEN) |_BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0)  ;  // Enable ADC, prescale at 128.

ADoutTemp = 0;


ADCSRA |= _BV(ADSC);	// Start initial ADC cycle



while ((ADCSRA & _BV(ADSC)) != 0){;;}		// Wait until conversion has finished

	ADIn = ADCW;	


if (ADIn < 975)
	charlieplexed = 0;
else
	charlieplexed = 1;
	











ADMUX = 3U;		// Channel 3 only

if (charlieplexed == 0) {

//Initialization for Multiplexed drive
DDRC  = 255; // All C Inputs
PORTC = 64U;

PORTA = 0;
PORTB = 0;
PORTC = 0;
PORTD = 0;

DDRA = 192U;
DDRB = 255U;
DDRC = 255U;
DDRD = 255U;

}
else
{


// Initial conditions for Charlieplexed drive
DDRD  = 0; // All D Inputs
PORTD = 0; // All D Low

DDRC  = 0; // All C Inputs
PORTC = 0;

DDRB  = 0; // All B Inputs
PORTB = 0;

DDRA  = 0; // All A Inputs
PORTA = 0;
}
		
		
asm("sei");		// ENABLE global interrupts




for (;;)  // main loop										
{




if ((ADCSRA & _BV(ADSC)) == 0)		// If conversion has finished
{	
	ADIn = ADCW;	
	
	
if (ADIn < 25)
{
	EnableLights = 1; 
	ADCcount = 0;
	AutoMode = 0; 
}	
else if (ADIn < 975)		// It's Dark out! Turn on the lights!
{

	if (EnableLights == 1)
		ADCcount++;
	else
		ADCcount = 0;
} 
else	// if (ADIn > 975), meaning that it's DARK out...
{
	if (EnableLights == 0)
		ADCcount++;
	else
		ADCcount = 0;
	
}
	
	if (ADCcount > 1000U){
	ADCcount = 0;
	AutoMode = 1;
	if (ADIn < 975)
	{
		EnableLights = 0;    
	}
	else{
		EnableLights = 1;  
 
	}	
	}

	ADCSRA |= _BV(ADSC);	// Start ADC
}






if (EnableLights) {

if (charlieplexed == 0) {

PORTA = 64;
delayLong();
PORTA = 128;
delayLong();
PORTA = 0; 

PORTB = 1;
delayLong();
PORTB = 2;
delayLong();
PORTB = 4;
delayLong();
PORTB = 8;
delayLong();
PORTB = 16;
delayLong();
PORTB = 32;
delayLong();
PORTB = 64;
delayLong();
PORTB = 128;
delayLong();
PORTB = 0; 

PORTC = 1;
delayLong();
PORTC = 2;
delayLong();
PORTC = 4;
delayLong();
PORTC = 8;
delayLong();
PORTC = 16;
delayLong();
PORTC = 32;
delayLong();
PORTC = 64;
delayLong();
PORTC = 128;
delayLong();
PORTC = 0; 

PORTD = 1;
delayLong();
PORTD = 2;
delayLong();
PORTD = 4;
delayLong();
PORTD = 8;
delayLong();
PORTD = 16;
delayLong();
PORTD = 32;
delayLong();
PORTD = 64;
delayLong();
PORTD = 128;
delayLong();
PORTD = 0; 
} // End Multiplexed Drive Section [if (charlieplexed == 0) {...}]
else
{ // Begin CharliePlexed drive section.  
  // NOTE: The microcontroller can only supply an aggregate maximum of 200 mA of current
  // at its power and ground pins at any given moment.  This limit must not be exceeded.
  // If using the usual 20 mA current per LED, NO MORE THAN NINE (9) LEDs may be on
  // at any given time.  So watch it.
  
  // Code that follows lights up ALL LEDs in a charlieplexed method, for use as a dim,
  // but energy efficient pegboard.
  



//
//
//PART A: Columns 0-7 (PORTD) high, taking EACH ROW low, one at a time.
//
//

DDRD  = 255; // All Outputs
PORTD = 255U;


//Take Rows 0-7 low, one at a time:

PORTD = 127; // All high but bit 7 (row 0)
dc();
PORTD = 191; // All high but bit 6 (row 1)
dc();
PORTD = 223; // All high but bit 5 (row 2)
dc();
PORTD = 239; // All high but bit 4 (row 3)
dc();
PORTD = 247; // All high but bit 3 (row 4)
dc();
PORTD = 251; // All high but bit 2 (row 5)
dc();
PORTD = 253; // All high but bit 1 (row 6)
dc();
PORTD = 254; // All high but bit 0 (row 7)
dc();

PORTD = 255; // All high  -> Now rows on.

//Take Rows 8-15 low, one at a time:

PORTC = 0; // All C Low
DDRC = 128; 
dc();
DDRC = 64;
dc();
DDRC = 32;
dc();
DDRC = 16;
dc();
DDRC = 8; 
dc();
DDRC = 4; 
dc();
DDRC = 2; 
dc();
DDRC = 1; 
dc();
DDRC = 0; // All inputs 

//Take Rows 16-24 low, one at a time:

PORTB = 0; // All Low
DDRB = 128; 
dc();
DDRB = 64;
dc();
DDRB = 32;
dc();
DDRB = 16;
dc();
DDRB = 8; 
dc();
DDRB = 4; 
dc();
DDRB = 2; 
dc();
DDRB = 1; 
dc();
DDRB = 0; // All inputs 


//Take Row 24 low:

PORTA = 0; // All Low
DDRA = 128; 
dc(); 
DDRA = 0; // All inputs 



DDRD  = 0; // All Outputs
PORTD = 0;

//END PART A (Columns 0-7 (PORTD) were high)

//
//
//PART B: Columns 0-7 (PORTC) high, taking EACH ROW low, one at a time.
//
//


DDRC  = 255; // All Outputs
PORTC = 255U;

//Take Rows 0-7 low, one at a time:

PORTD = 0; // All Low
DDRD = 128; 
dc();
DDRD = 64;
dc();
DDRD = 32;
dc();
DDRD = 16;
dc();
DDRD = 8; 
dc();
DDRD = 4; 
dc();
DDRD = 2; 
dc();
DDRD = 1; 
dc();
DDRD = 0; // All inputs 


//Take Rows 8-15 low, one at a time:

PORTC = 127; // All high but bit 7 (row 0)
dc();
PORTC = 191; // All high but bit 6 (row 1)
dc();
PORTC = 223; // All high but bit 5 (row 2)
dc();
PORTC = 239; // All high but bit 4 (row 3)
dc();
PORTC = 247; // All high but bit 3 (row 4)
dc();
PORTC = 251; // All high but bit 2 (row 5)
dc();
PORTC = 253; // All high but bit 1 (row 6)
dc();
PORTC = 254; // All high but bit 0 (row 7)
dc();

PORTC = 255; // All high  -> Now rows on.


//Take Rows 16-24 low, one at a time:

PORTB = 0; // All Low
DDRB = 128; 
dc();
DDRB = 64;
dc();
DDRB = 32;
dc();
DDRB = 16;
dc();
DDRB = 8; 
dc();
DDRB = 4; 
dc();
DDRB = 2; 
dc();
DDRB = 1; 
dc();
DDRB = 0; // All inputs 


//Take Row 24 low:

PORTA = 0; // All Low
DDRA = 128; 
dc(); 
DDRA = 0; // All inputs 



DDRC  = 0; // All Outputs
PORTC = 0;

//END PART B (Columns 8-15 (PORTC) were high)

//
//
//PART C: Columns 16-24 (PORTB, PORTA) high, taking EACH ROW low, one at a time.
//
//

DDRB  = 255; // All Outputs
PORTB = 255;

DDRA = 128;
PORTA = 128;


//Take Rows 0-7 low, one at a time:

PORTD = 0; // All Low
DDRD = 128; 
dc();
DDRD = 64;
dc();
DDRD = 32;
dc();
DDRD = 16;
dc();
DDRD = 8; 
dc();
DDRD = 4; 
dc();
DDRD = 2; 
dc();
DDRD = 1; 
dc();
DDRD = 0; // All inputs 


//Take Rows 8-15 low, one at a time:

PORTC = 0; // All C Low
DDRC = 128; 
dc();
DDRC = 64;
dc();
DDRC = 32;
dc();
DDRC = 16;
dc();
DDRC = 8; 
dc();
DDRC = 4; 
dc();
DDRC = 2; 
dc();
DDRC = 1; 
dc();
DDRC = 0; // All inputs 


//Take Rows 16-23 low, one at a time:

PORTB = 127; // All high but bit 7 (row 0)
dc();
PORTB = 191; // All high but bit 6 (row 1)
dc();
PORTB = 223; // All high but bit 5 (row 2)
dc();
PORTB = 239; // All high but bit 4 (row 3)
dc();
PORTB = 247; // All high but bit 3 (row 4)
dc();
PORTB = 251; // All high but bit 2 (row 5)
dc();
PORTB = 253; // All high but bit 1 (row 6)
dc();
PORTB = 254; // All high but bit 0 (row 7)
dc();


//Take Row 24 low:

PORTA = 0; // All Low
DDRA = 128; // Output pin 7
dc(); 
DDRA = 0; // All inputs 






DDRB  = 0; // All inputs
PORTB = 0;


//END PART C (Columns 16-24 (PORTB, PORTA) were high)

//BEGIN PART D ("Row" 25 high-- actually a diagonal *column*)

PORTA = 64U; // All Low
DDRA = 64U; // Output pin 7

PORTD = 0; // All Low
DDRD = 255U; 
dc();
DDRD = 0; 

PORTC = 0; // All Low
DDRC = 255U; 
dc();
DDRC = 0; 

PORTB = 0; // All Low
DDRB = 255U; 
dc();
DDRB = 0; 



} // End charlieplex section
} 


}	//End main loop.
	return 0;
}
